// clang-format off
/*
 * KDiff3 - Text Diff And Merge Tool
 *
 * SPDX-FileCopyrightText: 2021-2021 Michael Reeves reeves.87@gmail.com
 * SPDX-License-Identifier: GPL-2.0-or-later
 */
// clang-format on

#ifndef DIRECTORY_LIST_H
#define DIRECTORY_LIST_H

#include <list>

class FileAccess;
using DirectoryList =  std::list<FileAccess>;

#endif
